#Origionally modified from code by Jonathan Pekar
import json

#dict from https://pythonforbiologists.com/dictionaries
gencode = {
    'ATA':'I', 'ATC':'I', 'ATT':'I', 'ATG':'M',
    'ACA':'T', 'ACC':'T', 'ACG':'T', 'ACT':'T',
    'AAC':'N', 'AAT':'N', 'AAA':'K', 'AAG':'K',
    'AGC':'S', 'AGT':'S', 'AGA':'R', 'AGG':'R',
    'CTA':'L', 'CTC':'L', 'CTG':'L', 'CTT':'L',
    'CCA':'P', 'CCC':'P', 'CCG':'P', 'CCT':'P',
    'CAC':'H', 'CAT':'H', 'CAA':'Q', 'CAG':'Q',
    'CGA':'R', 'CGC':'R', 'CGG':'R', 'CGT':'R',
    'GTA':'V', 'GTC':'V', 'GTG':'V', 'GTT':'V',
    'GCA':'A', 'GCC':'A', 'GCG':'A', 'GCT':'A',
    'GAC':'D', 'GAT':'D', 'GAA':'E', 'GAG':'E',
    'GGA':'G', 'GGC':'G', 'GGG':'G', 'GGT':'G',
    'TCA':'S', 'TCC':'S', 'TCG':'S', 'TCT':'S',
    'TTC':'F', 'TTT':'F', 'TTA':'L', 'TTG':'L',
    'TAC':'Y', 'TAT':'Y', 'TAA':'_', 'TAG':'_',
    'TGC':'C', 'TGT':'C', 'TGA':'_', 'TGG':'W', "---": "-"}


input_seq_path = "/mnt/c/Users/hjenn/Documents/1_UCSD/Research/COVID/Muller_Lineage/15_updatedData/GISAID_NYCPHL_04_29/gisaid_2021_04_30_00_rename_aln.fasta"
outPrefixPath = "/mnt/c/Users/hjenn/Documents/1_UCSD/Research/COVID/Muller_Lineage/15_updatedData/GISAID_NYCPHL_04_29/renameAln_codingRegions"
annotation_path = '/mnt/c/Users/hjenn/Documents/1_UCSD/Research/COVID/epiMuller/data/genemap_SARSCoV2.gff'


fasta = {}
input_seq = open(input_seq_path, "r")
idLine = 'na'
for line in input_seq:
    if ">" in line:
        if idLine != "na":
            fasta[idLine] = seq
        idLine = line.strip()
        seq = ""
    else: 
        seq = seq + line.strip()
fasta[idLine] = seq

fasta_coding_d = {}
fasta_aa_d = {}
stop_codon_dict = {}
gene_boundry_d = {}
aa_index = 0

for l in open(annotation_path).readlines():
    if '#' in l:
        continue
    l = l.strip().split('\t')
    start = int(l[3])
    end = int(l[4])
    gene = l[-1].split('"')[1]
    if gene == 'ORF9b' or gene == 'ORF14': 
        continue
    

    #fasta_path = 'PATHTODFASTA_%s.fasta' % gene
    seq_set = set()
    seq_keys = []
    stop_codon_dict[gene] = {}
    
    #with open(fasta_path, 'w') as f:
    for k in fasta.keys():
        seq = fasta[k][start-1:end][:-3] # exclude stop codon

        seq_noStop = ''
        aaSeq_noStop = ""
        hadStop = False
        for i in range(int(len(seq)/3)):
            codon = seq[i*3:i*3+3].upper()
            if codon == 'TGA' or codon == 'TAA' or codon == 'TAG' or hadStop:
                hadStop = True
                seq_noStop += '---'
                aaSeq_noStop += "-"
            else:

                seq_noStop += codon
                if "N" in codon:
                    aa = "-"
                elif "-" in codon:
                    aa = "-"
                else:
                    if codon not in gencode:
                        aa = "-"
                    else:
                        aa = gencode[codon]
                aaSeq_noStop += aa

        #f.write('%s\n%s\n' % (k, seq_noStop)) 
        if k not in fasta_coding_d:
            fasta_coding_d[k] = ""
        fasta_coding_d[k] += seq_noStop

        if k not in fasta_aa_d:
            fasta_aa_d[k] = ""
        fasta_aa_d[k] += aaSeq_noStop

    gene_boundry_d[gene] = [aa_index, aa_index+int(len(seq)/3)]
    aa_index += int(len(seq)/3)


fasta_path = outPrefixPath + '_allCDS.fasta'
f = open(fasta_path, "w")
for k in fasta_coding_d:
    f.write('%s\n%s\n' % (k, fasta_coding_d[k])) 


fasta_aa_path = outPrefixPath + '_allAA.fasta'
f = open(fasta_aa_path, "w")
for k in fasta_aa_d:
    f.write('%s\n%s\n' % (k, fasta_aa_d[k])) 


json_path = outPrefixPath + '_geneAAboundries.json'
with open(json_path, 'w') as fp:
    json.dump(gene_boundry_d, fp)
    print("dumped json")






class parseTree:
	def __init__(self, root,  name="tree"):
		self.name = name
		self.root = root

class Node:
	def __init__(self, parent, length, name='internalNode'):
		self.parent = parent
		self.name = name
		self.length = length
		self.children = []
		self.annotKeys = []
		



import fnmatch
import os
import argparse
import sys
import re
from ete3 import Tree
import time
from datetime import timedelta, date
import csv
import json
from statistics import mode
import svgwrite #conda install svgwrite
import random
import cairo
import cairosvg



def main():


	MIN_PYTHON = (3, 7)
	if sys.version_info < MIN_PYTHON:
		sys.exit("Python %s.%s or later is required.\n" % MIN_PYTHON)

	print("Loaded imports")

	parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)


	##########################  parse user arguments
	parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)

	global_group  = parser.add_argument_group('Options for full repot')

	global_group.add_argument('-oDir', '--outDirectory', required=False, default ="./", type=str, help="folder for output")
	global_group.add_argument('-oP', '--outPrefix', required=True, type=str, help="prefix of out files withen outDirectory")


	defClades_group  = parser.add_argument_group('Options passed to epimuller-define')


	treeAndtraits = parser.add_mutually_exclusive_group(required=True)
	treeAndtraits.add_argument('-n', '--inNextstrain', type=str, help="nextstrain results with tree.nwk and [traitOfInterst].json")
	treeAndtraits.add_argument('-a', '--annotatedTree', type=str, help="nexus file name with annotation: [&!traitOfInterst=value], as output by treetime")

	defClades_group.add_argument('-m', '--inMeta', required=True, type=str, help="metadata tsv with 'strain' and 'date'cols, optional: cols of trait of interst; and pangolin col named:'pangolin_lineage', 'lineage' or 'pangolin_lin'")
	defClades_group.add_argument('-p', '--inPangolin', required=False, type=str, default = "metadata", help="pangolin output lineage_report.csv file, if argument not supplied looks in inMeta for col with 'pangolin_lineage', 'pangolin_lin', or 'lineage'")
	defClades_group.add_argument("--noPangolin", action="store_true", help="do not add lineage to clade names")

	defClades_group.add_argument('-k', '--traitOfInterstKey', required=False, type=str, default="aa_muts",  help="key for trait of interst in json file OR (if -a/--annotatedTree AND key is mutations with aa (not nuc):  use 'aa_muts')")
	defClades_group.add_argument('-f', '--traitOfInterstFile', required=False, type=str, default="aa_muts.json",  help="[use with -n/--inNextstrain] name of [traitOfInterstFile].json in '-n/--inNextstrain' folder")
	defClades_group.add_argument('-g', '--geneBoundry', required=False, type=str, help="[use with -a/--annotatedTree AND -k/--traitOfInterst aa_muts] json formated file specifing start end postions of genes in alignment for annotatedTree  (see example data/geneAAboundries.json)")

	defClades_group.add_argument('-mut', '--VOClist', required=False, nargs='+', help="list of aa of interest in form [GENE][*ORAncAA][site][*ORtoAA] ex. S*501*, gaps represed by X, wild card aa represented by *")

	defClades_group.add_argument('-t', '--timeWindow', required=False, type=str, default="7", help="number of days for sampling window")
	defClades_group.add_argument('-s', '--startDate', required=False, type=str, default='2020-03-01', help="start date in iso format YYYY-MM-DD or 'firstDate' which sets start date to first date in metadata")
	defClades_group.add_argument('-e', '--endDate', required=False, type=str, default='lastDate', help="end date in iso format YYYY-MM-DD or 'lastDate' which sets end date as last date in metadata")

	

	##########################  command line args for drawMuller.py

	drawing_group  = parser.add_argument_group('Options passed to epimuller-draw')

	drawing_group.add_argument('-mt', '--MINTIME', required=False, type=str, default="30", help="minimum time point to start plotting")
	drawing_group.add_argument('-min', '--MINTOTALCOUNT', required=False, type=str, default="50", help="minimum total count for group to be included")

	drawing_group.add_argument('-c', '--cases_name', required=False, type=str, help="file with cases - formated with 'date' in ISO format and 'confirmed_rolling' cases, in tsv format")
	drawing_group.add_argument("--avgWindow", required=False, type=str, help="width of rolling mean window in terms of --timeWindow's (recomend using with small --timeWindow) ; default: sum of counts withen timeWindow (ie no average)")


	drawing_group.add_argument('-l', '--xlabel', required=False, type=str, choices =  ["date", "time", "bimonthly"], default="date", help="Format of x axis label: ISO date format or timepoints from start, or dd-Mon-YYYY on 1st and 15th")
	drawing_group.add_argument('-lp', '--labelPosition', required=False, type=str, default="Right", choices = ["Right", "Max", "Start", "End"], help="choose position of clade labels")

	drawing_group_page  = parser.add_argument_group('Options passed to epimuller-draw for page setup')
	drawing_group_page.add_argument('--WIDTH', required=False, type=str, default="1500", help="WIDTH of page (px)")
	drawing_group_page.add_argument('--HEIGHT', required=False, type=str, default="1000", help="HEIGHT of page (px)")
	drawing_group_page.add_argument('--LEGENDWIDTH', required=False, type=str, default="220", help="LEGENDWIDTH to the right of plotting area (px)")
	drawing_group_page.add_argument('--MARGIN', required=False, type=str, default="60", help="MARGIN around all sides of plotting area (px)")
	drawing_group_page.add_argument('--FONTSIZE', required=False, type=str, default="26")
	drawing_group_page.add_argument('--LABELSHIFT', required=False, type=str, default="15", help="nudge label over by LABELSHIFT (px)")
	


	args = parser.parse_args()


	#call with script from testing folder
	# commandCallDefine = "python ../../epiMuller/scripts/defineAndCountClades.py"
	# commandCallDraw = "python ../../epiMuller/scripts/drawMuller.py"

	#call with script from epiMuller folder
	# commandCallDefine = "python ../scripts/defineAndCountClades.py"
	# commandCallDraw = "python ../scripts/drawMuller.py"

	#call with entry_points

	commandCallDefine = "epimuller-define"
	commandCallDraw = "epimuller-draw"


	########################### call defineAndCountClades.py

	if args.inNextstrain is not None:
		treeAndtraits = "--inNextstrain " + args.inNextstrain
	else:
		treeAndtraits = "--annotatedTree " + args.annotatedTree

	if args.noPangolin:
		noPangolin = " --noPangolin "
	else:
		noPangolin = ""

	if args.geneBoundry is not None:
		geneBoundry = " --geneBoundry " + args.geneBoundry + " "
	else:
		geneBoundry = ""



	oscommand = " ".join([commandCallDefine, "--outDirectory", args.outDirectory , "--outPrefix", args.outPrefix ,
	 treeAndtraits , "--inMeta", args.inMeta , "--inPangolin", args.inPangolin , 
	 "--traitOfInterstFile", args.traitOfInterstFile , "--traitOfInterstKey", args.traitOfInterstKey , geneBoundry + noPangolin + "--timeWindow", args.timeWindow ,
	 "--startDate", args.startDate ,"--endDate", args.endDate])


	if args.VOClist is not None:
		oscommand = oscommand + " --VOClist"
		for aa in args.VOClist:
			oscommand += " " + aa




	print("\n ### Call command ###")
	print(oscommand)
	print("\n")
	if os.system(oscommand) != 0:
		sys.exit("epimuller-define failed")


	##########################  call drawMuller.py


	outCladeHierarchy_name = os.path.join(args.outDirectory, args.outPrefix +"_defineCountClades", "cladeHierarchy.csv")
	outCounts_name = os.path.join(args.outDirectory, args.outPrefix +"_defineCountClades", "abundances.csv")
	plot_folder = os.path.join(args.outDirectory, args.outPrefix+ "_PLOTS", "")


	# oscommand = " ".join(["python 00_scripts/drawMuller.py --parentHierarchy_name", outCladeHierarchy_name, "--abundance_name", outCounts_name, "--outFolder", plot_folder,
	# 	"--xlabel", args.xlabel, "--labelPosition", args.labelPosition, "--MINTIME", args.MINTIME, "--MINTOTALCOUNT", 
	# 	args.MINTOTALCOUNT, "--xlabel", args.xlabel , "--labelPosition", args.labelPosition])
	oscommand = " ".join([commandCallDraw, "--parentHierarchy_name", outCladeHierarchy_name, "--abundance_name", outCounts_name, "--outFolder", plot_folder,
		"--xlabel", args.xlabel, "--labelPosition", args.labelPosition, "--MINTIME", args.MINTIME, "--MINTOTALCOUNT", 
		args.MINTOTALCOUNT, "--xlabel", args.xlabel , "--labelPosition", args.labelPosition, 
		"--WIDTH", args.WIDTH, "--HEIGHT", args.HEIGHT, "--LEGENDWIDTH", args.LEGENDWIDTH, "--MARGIN", args.MARGIN , "--FONTSIZE", args.FONTSIZE, "--LABELSHIFT", args.LABELSHIFT])

	if args.cases_name is not None:
		oscommand = oscommand + " --cases_name " + args.cases_name
	
	print("\n ### Call command ###")
	print(oscommand)
	print("\n")
	if os.system(oscommand) != 0:
		sys.exit("epimuller-draw failed")
	


if __name__ == "__main__":
	MIN_PYTHON = (3, 7)
	if sys.version_info < MIN_PYTHON:
		sys.exit("Python %s.%s or later is required.\n" % MIN_PYTHON)

	print("Loaded imports")
	main()


import fnmatch
import os
import argparse
import sys
import re
from ete3 import Tree
import time
from datetime import timedelta, date
import datetime
import csv
import json
from statistics import mode

import scripts.parse as parse



import random

def readInMeta(inMeta_name, pangolin):
	'''
	Input: ete3 tree with node names that have 'trait' of clade specified in j_d
	Outputs: tree with trait appened to node names
	'''
	firstDate = date.fromisoformat('2030-01-01')
	lastDate =  date.fromisoformat('1900-01-01')

	sampPangolin_d = {}
	sampDate_d = {}
	strain_index = "NA"
	pangolin_index = "NA"
	inMeta_open = open(inMeta_name, "r")
	for line in inMeta_open:
		line_l = re.split('\t|,', line.strip())
		if strain_index != "NA":
			tipName = line_l[strain_index]
			try:
				tipDate = line_l[date_index]
			except:
				tipDate = ""

			sampDate_d[tipName] = tipDate 

			if tipDate != "":
				try:
					isoDate = date.fromisoformat(tipDate)

					if isoDate < firstDate:
						firstDate = isoDate
					if isoDate > lastDate:
						lastDate = isoDate

				except ValueError:
					pass

			if pangolin == 'metadata' and pangolin_index != "NA":
				sampPangolin = line_l[pangolin_index]
				sampPangolin_d[tipName] = sampPangolin

		elif "strain" in line_l and "date" in line_l:
			strain_index = line_l.index("strain")
			date_index = line_l.index("date")

		elif "Name" in line_l and "Date" in line_l:
			strain_index = line_l.index("Name")
			date_index = line_l.index("Date")

		else:
			print("Metadata must have 'strain' and 'date' OR 'Name' and 'Date' in first line with tab or comma seperated and no spaces. Current col are:")
			sys.exit(line_l)
			#"Must have 'strain' and 'date' in first line of tab seperated"

		if pangolin == 'metadata' and pangolin_index == "NA":
			if "pangolin_lin" in line_l:
				pangolin_index = line_l.index("pangolin_lin")
			elif "pangolin_lineage" in line_l:
				pangolin_index = line_l.index("pangolin_lineage")
			elif "lineage" in line_l:
				pangolin_index = line_l.index("lineage")
			else:
				if sampDate_d == {}: #only print once
					print("pangolin not found in metadata, col must be called 'pangolin_lin', 'pangolin_lineage', or 'lineage' ")

	
	if pangolin != 'metadata': # pangolin output lineage report file name
		inPangolin_open = open(pangolin, "r")
		for line in inPangolin_open:
			line_l = line.split(",")
			if line_l[2] != "probability": #is not first line
				if float(line_l[2]) > 0.9:
					sampPangolin = line_l[1]
				else:
					print("Prob does not meet threshold", tipName, float(line_l[2]) )
					sampPangolin = 'NA'
				tipName = line_l[0]
				sampPangolin_d[tipName] = sampPangolin

	return(sampDate_d, sampPangolin_d, firstDate, lastDate)

def annotateNwk_nextstrain(t, j_d, trait, sampDate_d, sampPangolin_d, geneToIndex, indexToGene):

	for node in t.traverse("preorder"):
		if "NODE_" not in node.name:
			if node.name in sampPangolin_d:
				pID = sampPangolin_d[node.name]
			else: 
				pID = "NA"
			temp = "_" + pID + "_" + sampDate_d[node.name] + "_"
		else:
			temp = "_"

		if trait == "aa_muts":
			#for gene in geneToIndex:
			for index in range(len(indexToGene)):
				gene = indexToGene[index]
				if gene in j_d['nodes'][node.name][trait]:
					for mut in j_d['nodes'][node.name][trait][gene]:
						usable = mut.replace(".", "J").replace("-", "X").replace("*", "J") #gaps represented by X
						temp += usable+"."
				temp += "-"
		elif trait == "muts":
			for mut in j_d['nodes'][node.name][trait]:
				temp += mut + "."
		else:
			temp += str(j_d['nodes'][node.name][trait])
			#node_annot = temp
		node.name = node.name + temp
	return(t)


def treetimeToTraits_d(parseT, traitOfInterstKey):
	nodeTraits_d = {}
	if traitOfInterstKey == "aa_muts":
		traitOfInterstKey = "mutations"
	for node in parse.traverse_depthFirst(parseT.root):
		temp = getattr(node, traitOfInterstKey, None)
		if temp is not None:
			nodeTraits_d[node.name] = temp
		else:
			nodeTraits_d[node.name] = "NA"
		 	#print("does not have trait", node.__dict__)
	return nodeTraits_d


def annotateNwk_treetime(t, nodeTraits_d, trait, geneBoundry_d, sampDate_d, sampPangolin_d, geneToIndex, indexToGene):
	
	for node in t.traverse("preorder"):
		if "NODE_" not in node.name:
			if node.name in sampPangolin_d:
				pID = sampPangolin_d[node.name]
			else: 
				pID = "NA"
			temp = "_" + pID + "_" + sampDate_d[node.name] + "_"
		else:
			temp = "_"

		if trait == "aa_muts":
			mutList = str(nodeTraits_d[node.name])
			
			for index in range(12):
				refGene = indexToGene[index]
				if refGene in geneBoundry_d:
					for mutRaw in mutList.split(","):
						if len(mutRaw) > 2:
							rawPos = int(mutRaw[1:-1])
							if rawPos >= geneBoundry_d[refGene][0] and rawPos < geneBoundry_d[refGene][1]:
								mutRaw.replace(".", "J").replace("-", "X").replace("*", "J")
								temp += mutRaw[0] + str(rawPos - geneBoundry_d[refGene][0]) + mutRaw[-1] + "."
				temp += "-"
			temp += str(nodeTraits_d[node.name])
			#todo loop through and assign postion in appened aa to genes using renameAln_codingRegions_geneAAboundries.json
		elif trait == "mutations":
			for mut in nodeTraits_d[node.name].split(","):
				temp += mut + "."
		else:
			temp += str(nodeTraits_d[node.name])
		
			
		node.name = node.name + temp
	return(t)


def assignToSpecAA(t, mutList, logNotes_open, geneToIndex):
	'''
	input: annotated tree (t) and aa mut (mutList) to look for
	output:
	assignment_d: key: leaf node name; value: clade
	heierarchy (heiarchy_d: key:child clade; value:parent clade )
	'''
	clade_s = set()
	assignment_d = {}
	heiarchy_d = {}
	cladesAssinged_d ={}

	currentClade = 'anc'
	heiarchy_d['anc'] = 'NA'
	clade_s.add('anc')

	for node in t.traverse("preorder"):
		assignment_d[node.name] = currentClade

	for node in t.traverse("preorder"):

		node_name = node.name

		nodeMut_l = node_name.split("_")[-1].split("-")
		currentClade = assignment_d[node_name]

		mutOfInterst_count = 0

		for gene_and_mutToFind in mutList:
			for gene in geneToIndex:
				if gene == gene_and_mutToFind[:len(gene)]:
					mutToFind = gene_and_mutToFind[len(gene):]

					nodeMut_gene_catString = nodeMut_l[geneToIndex[gene]]
					nodeMut_gene_l = nodeMut_gene_catString.split(".")
					if nodeMut_gene_l != ['']:
						for nodeMut_one in nodeMut_gene_l:
							if nodeMut_one != '':

								if mutToFind[0] == "*" or mutToFind[0] == nodeMut_one[0]:
									if mutToFind[-1] == "*" or mutToFind[-1] == nodeMut_one[-1]:
										if mutToFind[1:-1] == nodeMut_one[1:-1]:
											mutOfInterst_count += 1
											mutFound = nodeMut_one
		if mutOfInterst_count == 1:
			if mutFound not in cladesAssinged_d:
				cladesAssinged_d[mutFound] = 0
			cladesAssinged_d[mutFound] += 1

			newClade = mutFound+ "_" + str(cladesAssinged_d[mutFound])
			clade_s.add(newClade)

			heiarchy_d[newClade] = currentClade
			assignment_d[node_name] = newClade

			for des_node in node.iter_descendants("postorder"):
				assignment_d[des_node.name] = newClade

		if mutOfInterst_count > 1:
			#print("found more than 1 mut of interst", node_name)
			logNotes_open.write("\n"+"found more than 1 mut of interst"+"\n")
			logNotes_open.write(node_name+"\n")



	return(assignment_d, heiarchy_d, clade_s)


def assignToNucMut(t, mutList, logNotes_open):
	'''
	input: annotated tree (t) and list of nucliotide mut (mutList) to look for
	output:
	assignment_d: key: leaf node name; value: clade
	heierarchy (heiarchy_d: key:child clade; value:parent clade )
	'''
	clade_s = set()
	assignment_d = {}
	heiarchy_d = {}
	cladesAssinged_d ={}

	currentClade = 'anc'
	heiarchy_d['anc'] = 'NA'
	clade_s.add('anc')

	for node in t.traverse("preorder"):
		assignment_d[node.name] = currentClade

	for node in t.traverse("preorder"):

		node_name = node.name

		nodeMut_l = node_name.split("_")[-1].split(".")
		currentClade = assignment_d[node_name]

		mutOfInterst_count = 0

		for mutToFind in mutList:
			for nodeMut_one in nodeMut_l:
				if mutToFind[1:-1] == nodeMut_one[1:-1]:
					if mutToFind[0] == "*" or mutToFind[0] == nodeMut_one[0]:
						if mutToFind[-1] == "*" or mutToFind[-1] == nodeMut_one[-1]:
							mutOfInterst_count += 1
							mutFound = nodeMut_one
		if mutOfInterst_count == 1:
			if mutFound not in cladesAssinged_d:
				cladesAssinged_d[mutFound] = 0
			cladesAssinged_d[mutFound] += 1

			newClade = mutFound+ "_" + str(cladesAssinged_d[mutFound])
			clade_s.add(newClade)

			heiarchy_d[newClade] = currentClade
			assignment_d[node_name] = newClade

			for des_node in node.iter_descendants("postorder"):
				assignment_d[des_node.name] = newClade

		if mutOfInterst_count > 1:
			#print("found more than 1 mut of interst", node_name)
			logNotes_open.write("\n"+"found more than 1 mut of interst"+"\n")
			logNotes_open.write(node_name+"\n")



	return(assignment_d, heiarchy_d, clade_s)


def assignToTraits(t, ofInterst_l = []):
	'''
	input: annotated tree (t) 
	Not currntly functional: if ofInterst_l is not empty only initialize when node has trait in least
	output:
	assignment_d: key: leaf node name; value: clade
	heierarchy (heiarchy_d: key:child clade; value:parent clade )
	'''
	clade_s = set()
	assignment_d = {}
	heiarchy_d = {}
	cladesAssinged_d ={}

	currentClade = 'anc'
	heiarchy_d['anc'] = 'NA'
	clade_s.add('anc')

	for node in t.traverse("preorder"):
		assignment_d[node.name] = currentClade

	for node in t.traverse("preorder"):

		if not node.is_root():

			node_name = node.name

			traitOfNode = node_name.split("_")[-1]
			currentClade = assignment_d[node_name]


			#define new calde because does not match parent
			if currentClade.split("_")[0] != traitOfNode and traitOfNode != "?":
				if traitOfNode not in cladesAssinged_d:
					cladesAssinged_d[traitOfNode] = 0
				cladesAssinged_d[traitOfNode] += 1
				newClade = traitOfNode + "_" + str(cladesAssinged_d[traitOfNode])
				clade_s.add(newClade)

				heiarchy_d[newClade] = currentClade
				assignment_d[node_name] = newClade

				for des_node in node.iter_descendants("postorder"):
					assignment_d[des_node.name] = newClade


	return(assignment_d, heiarchy_d, clade_s)


def assignCladeToLin(assignment_old_d, heiarchy_old_d, clade_old_s):
	"""
	mode of Pangolin linage of tips withen clade are appended to clade names
	"""


	reassignCladeNames ={}
	clade_s = set()
	assignment_d = {}
	heiarchy_d = {}
	
	for clade in clade_old_s:
		tipLinForClade = []
		for node in assignment_old_d:
			if 'NODE' not in node:
				if assignment_old_d[node] == clade:
					#print(node)
					tipLinForClade = tipLinForClade + [node.split("_")[-3]]
		if clade != 'anc':
			try:
				lineage = mode(tipLinForClade)
			except ValueError:
				lineage = "2lin"
			newClade = lineage + "_" + clade
		else:
			newClade = clade

		
		reassignCladeNames[clade] = newClade
		clade_s.add(newClade)

	for child in heiarchy_old_d:
		parent = heiarchy_old_d[child]
		if parent != 'NA':

			heiarchy_d[reassignCladeNames[child]] = reassignCladeNames[parent]
		else:
			heiarchy_d[reassignCladeNames[child]] = parent
	for samp in assignment_old_d:
		assignment_d[samp] = reassignCladeNames[assignment_old_d[samp]]

	return (assignment_d, heiarchy_d, clade_s)


def countAbudanceFromNames_byWeek(assignment_d, clade_s, startDate, endDate, delta, tipLog_name):
	"""
	counts total number of tips withen each clade, for each time interval (delta) between startDate and endDate 
	"""
	psuodocount = 0.1
	abundances_d = {} # key: week; value: dict of key:clade; value: count
	weekToDate_d = {}
	#assignment_d: key: node name; value: clade

	week_l = []

	tipLog_open = open(tipLog_name, "w")
	outLine = "	".join(["Week", "Clade", "sample_withAnnot", "sample_inputID"]) + "\n"
	tipLog_open.write(outLine)

	currentStart = startDate
	currentEnd = currentStart + delta
	week = 0
	weekName = str(week)
	week_l = week_l + [weekName]
	abundances_d[weekName] = {}

	while endDate >= currentEnd:

		lastWeekName = weekName
		week += 1
		weekName = str(week)
		week_l = week_l + [weekName]

		weekToDate_d[weekName] = currentStart

		abundances_d[weekName] = {} 

		for clade in abundances_d[lastWeekName]:
			abundances_d[weekName][clade] = psuodocount #change to add psuodocount
			# if clade == 'anc':
			# 	abundances_d[weekName][clade] += 1


		for tip in assignment_d.keys():
			if "NODE_" not in tip and "Wuhan" not in tip:
				try: 
					tip_date = date.fromisoformat(tip.split("_")[-2])

					if tip_date < currentEnd and tip_date >= currentStart:

						clade = assignment_d[tip]
						if clade not in abundances_d[weekName]:
							abundances_d[weekName][clade] = psuodocount #change to add psuodocount
						abundances_d[weekName][clade] += 1
						if clade != 'anc':
							tipNoAnnot = ""
							for name in tip.split("_")[:-3]: 
								tipNoAnnot = tipNoAnnot + "_"  + name
							outLine = "	".join([ weekName, clade, tip, tipNoAnnot[1:]]) + "\n"
							tipLog_open.write(outLine)
				except ValueError:
					pass 


		currentEnd += delta
		currentStart += delta
	tipLog_open.close()

	noFurtherAbudance = list(clade_s.copy())
	for weekName in (reversed(week_l)):

		noFurtherAbudance_last = noFurtherAbudance.copy()
		for clade in noFurtherAbudance_last:
			if clade in abundances_d[weekName]:

				if abundances_d[weekName][clade] == psuodocount: #and clade != 'anc':
					abundances_d[weekName][clade] = 0
				else:
					noFurtherAbudance.remove(clade)

	return(abundances_d, weekToDate_d)


def main():



	##########################  parse user arguments
	parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)

	parser.add_argument('-oDir', '--outDirectory', required=False, default ="./", type=str, help="folder for output")
	parser.add_argument('-oP', '--outPrefix', required=True, type=str, help="prefix of out files withen outDirectory")


	treeAndtraits = parser.add_mutually_exclusive_group(required=True)
	treeAndtraits.add_argument('-n', '--inNextstrain', type=str, help="nextstrain results with tree.nwk and [traitOfInterstFile].json")
	treeAndtraits.add_argument('-a', '--annotatedTree', type=str, help="nexus file name with annotation: [&!traitOfInterstKey=value], as output by treetime")


	parser.add_argument('-m', '--inMeta', required=True, type=str, help="metadata tsv with 'strain'	and 'date'cols, optional: col for  [traitOfInterstKey]; and pangolin col named: 'pangolin_lineage' 'lineage' or 'pangolin_lin'")
	parser.add_argument('-p', '--inPangolin', required=False, type=str, default = "metadata", help="pangolin output lineage_report.csv file, if argument not supplied looks in inMeta for col with 'pangolin_lineage', 'pangolin_lin', or 'lineage'")
	parser.add_argument("--noPangolin", action="store_true", help="do not add lineage to clade names")


	parser.add_argument('-k', '--traitOfInterstKey', required=False, type=str, default="aa_muts",  help="key for trait of interst in json file OR (if -a/--annotatedTree AND key is mutations with aa (not nuc):  use 'aa_muts')")
	parser.add_argument('-f', '--traitOfInterstFile', required=False, type=str, default="aa_muts.json",  help="[use with -n/--inNextstrain] name of [traitOfInterstFile].json in '-n/--inNextstrain' folder")
	parser.add_argument('-g', '--geneBoundry', required=False, type=str, help="[use with -a/--annotatedTree AND -k/--traitOfInterst aa_muts] json formated file specifing start end postions of genes in alignment for annotatedTree  (see example data/geneAAboundries.json)")
	

	parser.add_argument('-mut', '--VOClist', required=False, nargs='+', help="list of aa of interest in form [GENE][*ORAncAA][site][*ORtoAA] ex. S*501*, gaps represented by X, wild card aa represented by *")

	parser.add_argument('-t', '--timeWindow', required=False, type=str, default="7", help="number of days for sampling window")
	parser.add_argument('-s', '--startDate', required=False, type=str, default='2020-03-01', help="start date in iso format YYYY-MM-DD or 'firstDate' which is in metadata")
	parser.add_argument('-e', '--endDate', required=False, type=str, default='lastDate', help="end date in iso format YYYY-MM-DD or 'lastDate' which is in metadata")

	

	args = parser.parse_args()

	##########################  file names

	
	outCladeHierarchy_name = os.path.join(args.outDirectory, args.outPrefix +"_defineCountClades", "cladeHierarchy.csv")
	outCounts_name = os.path.join(args.outDirectory, args.outPrefix +"_defineCountClades", "abundances.csv")
	tempTree_name = os.path.join(args.outDirectory, args.outPrefix +"_defineCountClades", "tree.nwk")
	tipLog_name = os.path.join(args.outDirectory, args.outPrefix +"_defineCountClades", "tipsWithMut.txt")

	logNotes_name = os.path.join(args.outDirectory, args.outPrefix + "_defineCountClades", "log.txt")


	if not os.path.exists(args.outDirectory):
		os.makedirs(args.outDirectory)

	dataOutDir = os.path.join(args.outDirectory, args.outPrefix + "_defineCountClades")
	if not os.path.exists(dataOutDir):
	 	os.makedirs(dataOutDir)


	######################### params to referance


	delta= timedelta(days=int(args.timeWindow))


	#########################  load treeAndTraits files

	sampDate_d, sampPangolin_d, firstDate, lastDate = readInMeta(args.inMeta, args.inPangolin)
	
	#use nextstrain as input
	if args.inNextstrain is not None:

		inTree_name = os.path.join(args.inNextstrain, "tree.nwk")
		print(inTree_name)
		if not os.path.exists(inTree_name):
			sys.exit("missing input tree")
		t = Tree(inTree_name, format = 3)


		#annotate tree with ancestral reconstruction
		inJSON_name = os.path.join(args.inNextstrain, args.traitOfInterstFile)
		j_d = json.load(open(inJSON_name))
		
		# extract all gene names
		if args.traitOfInterstKey == "aa_muts":
			i = 0
			geneToIndex_temp = {}
			indexToGene_temp = {}
			for node in j_d['nodes'].keys():
				for gene in j_d['nodes'][node]["aa_muts"].keys():
					if gene not in geneToIndex_temp:
						geneToIndex_temp[gene] = i
						indexToGene_temp[i] = gene
						i += 1
			# global geneToIndex
			geneToIndex = geneToIndex_temp.copy()
			# global indexToGene
			indexToGene = indexToGene_temp.copy()
		else:
			geneToIndex = {}
			indexToGene = {}

		t = annotateNwk_nextstrain(t, j_d, args.traitOfInterstKey, sampDate_d, sampPangolin_d, geneToIndex, indexToGene)
	#use treetime ancestral as input
	else: 
		parseT = parse.treeImport_wrap(args.annotatedTree, ["nexus", "treetimeAnnot"])
		print(tempTree_name)
		parse.writeNewick(parseT, tempTree_name, []) #write to file that can be read by ete3
		t = Tree(tempTree_name, format = 3)

		#annotate tree with ancestral reconstruction
		if args.traitOfInterstKey=="aa_muts":
			# extract all gene names
			if args.geneBoundry is None :
				sys.exit("geneBoundry json file is required for annotatedTree with aa_muts")
			else:
				geneBoundry_d = json.load(open(args.geneBoundry))

				i = 0
				geneToIndex_temp = {}
				indexToGene_temp = {}
				for gene in geneBoundry_d.keys():
					if gene not in geneToIndex_temp:
						geneToIndex_temp[gene] = i
						indexToGene_temp[i] = gene
						i += 1
				#global geneToIndex
				geneToIndex = geneToIndex_temp.copy()
				#global indexToGene
				indexToGene = indexToGene_temp.copy()
		else:
			geneBoundry_d = {}
			geneToIndex = {}
			indexToGene = {}


		nodeTraits_d = treetimeToTraits_d(parseT, args.traitOfInterstKey)
		t = annotateNwk_treetime(t, nodeTraits_d, args.traitOfInterstKey, geneBoundry_d,  sampDate_d, sampPangolin_d, geneToIndex, indexToGene)


	######################### make clades assignments (assignment_d: key: leaf node name; value: clade) and heierarchy (heiarchy_d: key:child clade; value:parent clade )

	if args.startDate == "firstDate":
		startDate = firstDate
	else:
		startDate =  date.fromisoformat(args.startDate)
	if args.endDate == "lastDate":
		endDate = lastDate
	else:
		endDate =  date.fromisoformat(args.endDate)	


	logNotes_open = open(logNotes_name, "w")


	if args.traitOfInterstKey == 'aa_muts':
		if args.VOClist is None:
			aa_mut = ["S*484*", "S*501*", "S*13*", "S*452*", "S*477*", "N*205*", "S*253*"]
		else:
			aa_mut = args.VOClist

		print("\n")
		print("Searching for aa muations with:", aa_mut)
		logNotes_open.write("\n"+"Searching for aa muations with:"+str(aa_mut)+"\n")
		assignment_d, heiarchy_d, clade_s = assignToSpecAA(t, aa_mut, logNotes_open, geneToIndex)
	elif args.traitOfInterstKey == 'mutations' or args.traitOfInterstKey == 'muts':
		if args.VOClist is None:
			nuc_mut = ["*23063*", "*22320*"] #mostly for testing purposes 
		else:
			nuc_mut = args.VOClist

		print("\n")
		print("Searching for muations with:", nuc_mut)
		logNotes_open.write("\n"+"Searching for muations with:"+str(nuc_mut)+"\n")
		assignment_d, heiarchy_d, clade_s = assignToNucMut(t, nuc_mut, logNotes_open)
	else:
		print("\n"+"Finding clades defined by"+str(args.traitOfInterstKey))
		logNotes_open.write("\n"+"Finding clades defined by"+str(args.traitOfInterstKey)+"\n")
		assignment_d, heiarchy_d, clade_s = assignToTraits(t)



	if sampPangolin_d != {} and not args.noPangolin:
		assignment_d, heiarchy_d, clade_s = assignCladeToLin(assignment_d, heiarchy_d, clade_s)


	######################### count abundances based on assignment_d


	abundances_d, weekToDate_d = countAbudanceFromNames_byWeek(assignment_d, clade_s, startDate, endDate, delta, tipLog_name)

	#abundances_d, heiarchy_d = clipPreParent(abundances_d, heiarchy_d)


	# ##########################  write abundances over time

	outCounts_open = open(outCounts_name, "w")
	headLine = "names,times,abundances,date\n"
	outCounts_open.write(headLine)

	for week in abundances_d:
		for name in abundances_d[week]:
			outLine = ",".join([name, str(week.replace("week", "")), str(abundances_d[week][name]), str(weekToDate_d[week])]) + "\n"
			outCounts_open.write(outLine)
	outCounts_open.close()

	##########################  write CladeHierarchy to file

	outCladeHierarchy_open = open(outCladeHierarchy_name, "w")
	headLine = "names,parents \n"
	outCladeHierarchy_open.write(headLine)
	for child in heiarchy_d:
		outLine = child + "," + heiarchy_d[child] + "\n"
		outCladeHierarchy_open.write(outLine)

	outCladeHierarchy_open.close()


	logNotes_open.write("\n"+"Final heiarchy_d"+"\n")
	logNotes_open.write(str(heiarchy_d)+"\n")


if __name__ == "__main__":
	MIN_PYTHON = (3, 7)
	if sys.version_info < MIN_PYTHON:
		sys.exit("Python %s.%s or later is required.\n" % MIN_PYTHON)

	print("Loaded imports")
	main()



import fnmatch
import os
import argparse
import sys
import re

from scripts.treeClass import Node, parseTree

def getTreeRawString(tree_fileName, treeformat):
	'''Extracts string that defines free from newick or nexus file'''

	treeformat_d = parseFormat(treeformat)

	tree_fileOpen = open(tree_fileName, "r")

	if treeformat_d["fileFormat"] == "nexus":
		nextLine = False

		for line in tree_fileOpen:
			if 'Begin Trees;' in line or "begin trees" in line:
				nextLine = True
			elif nextLine:
				#treeRawString = "=".join(line.strip().split("=")[1:])
				temp = line.strip()
				treeIndexStart = temp.index("(")
				treeRawString = temp[treeIndexStart:]
				nextLine = False
	elif treeformat_d["fileFormat"] == "newick":
		treeRawString = ""
		for line in tree_fileOpen:
			treeRawString += line.strip()
	else:
		"do not understand tree fileFormat"
		sys.exit()
 
	tree_fileOpen.close()

	if treeRawString[-1] == ";":
		treeRawString = treeRawString[:-1]

	return treeRawString

def matchParenCommas(treeString):
	'''
	input string
	output dictionaries of openParen_d: key:index of open parentheses, value: index of close parentheses; 
						   closeParn_d  key:index of close parentheses, value: index of open parentheses; 
						   commaParn_d:  key:index of open parentheses, value: list of indices of commas in that parenthesis levels
	'''
	stack_paren = []
	stack_comma = []
	openParen_d = {}
	closeParn_d = {}
	commaParn_d = {}
	notInAnnot = True

	for i, c in enumerate(treeString):
		if c == "(":
			stack_paren.append(i)
			stack_comma.append([])
		elif c == "," and notInAnnot:
			stack_comma[-1].append(i)
		elif c == ")":
			openParen = stack_paren.pop()
			openParen_d[openParen] = i
			closeParn_d[i] = openParen

			commaList = stack_comma.pop()
			commaParn_d[openParen] = commaList
		elif c == "[" or  c == "]" or c == "{" or c == "}":
			notInAnnot = not notInAnnot

	return(openParen_d, closeParn_d, commaParn_d)



def parseFormat(treeformat):
	'''
	checks for proporties of tree format list, that deviate from default; and returns dictionary
	'''
	
	treeformat_d = {"fileFormat":"newick", "branchLen":True, "internalNames":True, "leafNames":True, "annotation":False, "annotationType":None}

	if "nexus" in treeformat:
		treeformat_d["fileFormat"] = "nexus"


	if "noBranchLen" in treeformat:
		treeformat_d["branchLen"] = False


	if "treetimeAnnot" in treeformat:
		treeformat_d["annotation"] = True
		treeformat_d["annotationType"] = "treetimeAnnot"
	elif "hyphyAnnot" in treeformat:
		treeformat_d["annotation"] = True
		treeformat_d["annotationType"] = "hyphyAnnot"

	return (treeformat_d)


def buildTree(treeString, treeformat, openParen_d, closeParn_d, commaParn_d, subTstart, subTend, parent):
	'''
	builds tree by nodes pointing to parent and child based on input tree tree string
	returns root Node object
	'''


	currentNode = Node(parent, None, None)
	currntTreeString = treeString[subTstart+1:subTend]



	if "(" in currntTreeString: 
		#internal nodes

		firstParen = min(openParen_d.keys())
		closeParen = openParen_d.pop(firstParen)
		slitCommas = commaParn_d[firstParen]

		currentNodeString = treeString[closeParen+1:subTend]
		
		allSplits = [firstParen] + slitCommas + [closeParen]
		for i in range(len(allSplits)-1):
			childNode = buildTree(treeString, treeformat, openParen_d, closeParn_d, commaParn_d, allSplits[i], allSplits[i+1], currentNode)
			currentNode.children.append(childNode)

	else:
		#leaf
		currentNodeString = currntTreeString


	# parse currentNodeString name for details
	treeformat_d = parseFormat(treeformat)

	if treeformat_d["branchLen"]:
		if ":" in currentNodeString:
			currentNodeLen = currentNodeString.split(":")[1]		
			currentNodeName = currentNodeString.split(":")[0]
		else: #should only be in root for unrooted tree
			currentNodeLen = str(0)
			currentNodeName = currentNodeString
	else:
		currentNodeLen = str(1)
		currentNodeName = currentNodeString

	if treeformat_d["annotation"]:
		if treeformat_d["annotationType"] == "hyphyAnnot":
			annotation_l = currentNodeName.split("{")[-1].split(",")
			setattr(currentNode, "labels_l", annotation_l)
			currentNode.annotKeys.append("labels_l")
			currentNodeName = currentNodeName.split("{")[0]

		elif treeformat_d["annotationType"] == "treetimeAnnot":
			if "[" in str(currentNodeLen): #note: no annotation assigned to root 

				annotation_str = currentNodeLen.split("[")[1].replace("&", "")
				annotKey = ""
				annotValue = ""
				addToKey = True
				outOfQuote = True
				annotation_d = {}
				for c in annotation_str:
					if c == '"':
						outOfQuote = not outOfQuote

					if c == "=":
						addToKey = False
					elif (c == "," and outOfQuote) or c == "]":
						addToKey = True
						setattr(currentNode, annotKey, annotValue.replace('"', ''))
						currentNode.annotKeys.append(annotKey)
						annotKey = ""
						annotValue = ""
					elif addToKey:
						annotKey += c
					else:
						annotValue += c


			currentNodeLen = currentNodeLen.split("[")[0]

	currentNode.name = currentNodeName
	currentNode.length = currentNodeLen

	
	return currentNode


def traverse_depthFirst(currentNode):
	'''
	interates through nodes desendant from currentNode
	'''

	nodeList = [currentNode]

	for child in currentNode.children:
		nodeList += traverse_depthFirst(child)

	#print(nodeList)
	return nodeList



def treeImport_wrap(treefile, inTreeFormat):
	'''
	imports a tree from file and returns parseTree object
	'''
	treeRawString = getTreeRawString(treefile, inTreeFormat)
	openParen_d, closeParn_d, commaParn_d = matchParenCommas(treeRawString)
	root  = buildTree(treeRawString, inTreeFormat, openParen_d, closeParn_d, commaParn_d, 0, len(treeRawString), None)
	t = parseTree(root)

	return t



def makeNewickString(node, treeformat_d):
	'''
	helper for writeNewick which takes current node and recusively builds (chlidren newick)  then appends name as specified in treeformat_d
	'''
	if node.children != []:
		treeString = "("

		for child in node.children:
			childNwk = makeNewickString(child, treeformat_d)
			treeString += childNwk + ","

		treeString = treeString[:-1] + ")"
	else:
		treeString = ""

	treeString += node.name

	if treeformat_d["annotationType"] == "hyphyAnnot":
		treeString += "{"
		for label in node.labels_l:
			treeString += label + ","
		treeString = treeString[:-1] + "}"

	if treeformat_d["branchLen"]:
		treeString += ":" + str(node.length)
		if treeformat_d["annotationType"] == "hyphyAnnot":
			treeString += "[&"
			for annotKey in node.annotKeys:
				annotValue = node.annotKey
				treeString += annotKey + "=" + annotValue + ","
			treeString = treeString[:-1] + "]"



	return treeString


def writeNewick(tree, filename, treeformat):
	'''
	takes a parseTree object and writes in filenames a tree based on specifications of treeformat
	'''

	treeformat_d = parseFormat(treeformat)
	treeString = makeNewickString(tree.root, treeformat_d) + ";"

	directory = os.path.dirname(filename)
	if not os.path.exists(directory):
		os.makedirs(directory)

	file_open = open(filename, "w")
	file_open.write(treeString)
	file_open.close()




	# print("**************Tree: depthfirst*********")
	# for node in traverse_depthFirst(t.root):
	# 	print(node.name, node.length, node.date)#, "date:", node.date, ":", node.l)


